/*
 * ++Copyright++ 1985, 1989, 1993
 * -
 * Copyright (c) 1985, 1989, 1993
 *    The Regents of the University of California.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by the University of
 * 	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * -
 * Portions Copyright (c) 1993 by Digital Equipment Corporation.
 * 
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies, and that
 * the name of Digital Equipment Corporation not be used in advertising or
 * publicity pertaining to distribution of the document or software without
 * specific, written prior permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND DIGITAL EQUIPMENT CORP. DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL DIGITAL EQUIPMENT
 * CORPORATION BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * -
 * --Copyright--
 */

#if defined(LIBC_SCCS) && !defined(lint)
static char Wsccsid[] = "@(#)res_send.c	2.7 (motonori/WIDE) 3/10/95";
static char _sccsid[] = "@(#)res_send.c	8.1 (Berkeley) 6/4/93";
static char _rcsid[] = "$Id: res_send.c,v 4.9.1.12 1994/06/26 04:00:18 vixie Exp $";
#endif /* LIBC_SCCS and not lint */

	/* change this to "0"
	 * if you talk to a lot
	 * of multi-homed SunOS
	 * ("broken") name servers.
	 */
#define	CHECK_SRVR_ADDR	1	/* XXX - should be in options.h */

/*
 * Send query to name server and wait for reply.
 */

#include "sendmail.h"
#include <sys/uio.h>
#include <errno.h>
#include <arpa/nameser.h>
#include <resolv.h>


#ifndef INT16SZ
#define INT16SZ	2		/* for systems without 16-bit ints */
#endif
#ifndef HFIXEDSZ
#define HFIXEDSZ	12		/* #/bytes of fixed data in header */
#endif
#ifndef RES_INSECURE1
#define RES_INSECURE1	0x00000400	/* type 1 security disabled */
#endif
#ifndef RES_INSECURE2
#define RES_INSECURE2	0x00000800	/* type 2 security disabled */
#endif
#ifndef RES_NONBLOCK
#define RES_NONBLOCK	0x00001000	/* return immediately, do cacheing */
#endif

void _res_close ();

static int s = -1;	/* socket used for communications */
static int connected = 0;	/* is the socket connected */
static int vc = 0;	/* is the socket a virtual ciruit? */

#ifndef FD_SET
/* XXX - should be in portability.h */
#define	NFDBITS		32
#define	FD_SETSIZE	32
#define	FD_SET(n, p)	((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define	FD_CLR(n, p)	((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define	FD_ISSET(n, p)	((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#define FD_ZERO(p)	bzero((char *)(p), sizeof(*(p)))
#endif

#ifndef DEBUG
#   define Dprint(cond, args) /*empty*/
#   define DprintQ(cond, args, query) /*empty*/
#   define Aerror(file, string, error, address) /*empty*/
#   define Perror(file, string, error) /*empty*/
#else
#   define Dprint(cond, args) if (cond) {fprintf args;} else {}
#   define DprintQ(cond, args, query) if (cond) {\
			fprintf args;\
			__p_query(query);\
		} else {}
    static void
    Aerror(file, string, error, address)
	FILE *file;
	char *string;
	int error;
	struct sockaddr_in address;
    {
	int save = errno;

	if (_res.options & RES_DEBUG) {
		fprintf(file, "res_send: %s ([%s].%d): %s\n",
			string,
			inet_ntoa(address.sin_addr),
			address.sin_port,
			strerror(error));
	}
	errno = save;
    }
    static void
    Perror(file, string, error)
	FILE *file;
	char *string;
	int error;
    {
	int save = errno;

	if (_res.options & RES_DEBUG) {
		fprintf(file, "res_send: %s: %s\n",
			string, strerror(error));
	}
	errno = save;
    }
#endif

#ifdef notdef
static res_send_qhook Qhook = NULL;
static res_send_rhook Rhook = NULL;

void
res_send_setqhook(hook)
	res_send_qhook hook;
{
	Qhook = hook;
}

void
res_send_setrhook(hook)
	res_send_rhook hook;
{
	Rhook = hook;
}
#endif

/* int
 * our_server(ina)
 *	looks up "ina" in _res.ns_addr_list[]
 * returns:
 *	0  : not found
 *	>0 : found
 * author:
 *	paul vixie, 29may94
 */
static int
our_server(inp)
	struct sockaddr_in *inp;
{
	struct sockaddr_in ina;
	register int ns, ret;

	ina = *inp;
	ret = 0;
	for (ns = 0;  ns < _res.nscount;  ns++) {
		register struct sockaddr_in *srv = &_res.nsaddr_list[ns];

		if (srv->sin_family == ina.sin_family &&
		    srv->sin_port == ina.sin_port &&
		    (srv->sin_addr.s_addr == INADDR_ANY ||
		     srv->sin_addr.s_addr == ina.sin_addr.s_addr)) {
			ret++;
			break;
		}
	}
	return (ret);
}

/* int
 * name_in_query(name, type, class, buf, eom)
 *	look for (name,type,class) in the query section of packet (buf,eom)
 * returns:
 *	-1 : format error
 *	0  : not found
 *	>0 : found
 */
static int
name_in_query(name, type, class, buf, eom)
	char *name;
	register int type, class;
	u_char *buf, *eom;
{
	register u_char *cp = buf + HFIXEDSZ;
	int qdcount = ntohs(((HEADER*)buf)->qdcount);

	while (qdcount-- > 0) {
		char tname[MAXDNAME+1];
		register int n, ttype, tclass;

		n = dn_expand(buf, eom, cp, (u_char *)tname, sizeof tname);
		if (n < 0)
			return (-1);
		cp += n;
		ttype = _getshort(cp);	cp += INT16SZ;
		tclass = _getshort(cp); cp += INT16SZ;
		if (ttype == type &&
		    tclass == class &&
		    strcasecmp(tname, name) == 0)
			return (1);
	}
	return (0);
}

/* int
 * queries_match(buf1, eom1, buf2, eom2)
 *	is there a 1:1 mapping of (name,type,class)
 *	in (buf1,eom1) and (buf2,eom2)?
 * returns:
 *	-1 : format error
 *	0  : not a 1:1 mapping
 *	>0 : is a 1:1 mapping
 */
static int
queries_match(buf1, eom1, buf2, eom2)
	u_char *buf1, *eom1;
	u_char *buf2, *eom2;
{
	register u_char *cp = buf1 + HFIXEDSZ;
	int qdcount = ntohs(((HEADER*)buf1)->qdcount);

	if (qdcount != ntohs(((HEADER*)buf2)->qdcount))
		return (0);
	while (qdcount-- > 0) {
		char tname[MAXDNAME+1];
		register int n, ttype, tclass;

		n = dn_expand(buf1, eom1, cp, (u_char *)tname, sizeof tname);
		if (n < 0)
			return (-1);
		cp += n;
		ttype = _getshort(cp);	cp += INT16SZ;
		tclass = _getshort(cp); cp += INT16SZ;
		if (!name_in_query(tname, ttype, tclass, buf2, eom2))
			return (0);
	}
	return (1);
}

#if RES_NONBLOCK
struct cache {
	u_char	*request;
	u_char	*answer;
	int	reqlen;
	int	anslen;
	u_char	try;
	u_char	ns;
	u_char	badns;
	u_char	nocache;
	time_t	time;
	struct cache *next;
};

static struct cache *Caches = NULL;
#endif

int
res_send(buf, buflen, ans, anssiz)
#if defined(__RES)
#if (__RES > 19940221)	/* 4.9.3 */
	const u_char	*buf;
	int		buflen;
	u_char		*ans;
	int		anssiz;
#else			/* 4.9.2 */
	const char	*buf;
	int		buflen;
	char		*ans;
	int		anssiz;
#endif
#else			/* 4.8 */
	char		*buf;
	int		buflen;
	char		*ans;
	int		anssiz;
#endif
{
	HEADER		*hp = (HEADER *) buf;
	HEADER		*anhp = (HEADER *) ans;
	int		gotsomewhere = 0,
			connreset = 0,
			terrno = ETIMEDOUT;
#if RES_NONBLOCK
	struct cache *creq, *pcreq;
#endif
	register int	n;
	int		try, v_circuit, resplen, ns;
	u_int		badns;	/* XXX NSMAX can't exceed #/bits in this var */

	DprintQ((_res.options & RES_DEBUG) || (_res.pfcode & RES_PRF_QUERY),
		(stdout, ";; res_send()\n"), buf);
	if (!(_res.options & RES_INIT)) {
		if (res_init() == -1)
			return (-1);
	}
#if RES_NONBLOCK
	if (_res.options & RES_NONBLOCK) {
		if (LogLevel > 10)
			syslog(LOG_DEBUG, "Goto res_cache\n");
		/* goto caching query routine */
		return res_cache(buf, buflen, ans, anssiz);
	}
	pcreq = NULL;
	/* scan the query in cache */
	for (creq = Caches; creq != NULL; pcreq = creq, creq = creq->next)
	{
		if (creq->request != NULL
		 && queries_match(buf, buf + buflen, creq->request,
				  creq->request+creq->reqlen) > 0)
			break; /* already registed */
	}
	if (creq != NULL && creq->nocache == 0) {	/* found in cache */
		if (creq->answer != NULL && creq->anslen <= anssiz) {
			if (LogLevel > 10)
				syslog(LOG_DEBUG, "res_send: found in cache\n");
			if (creq->time < time(NULL)) {
				/* TTL expiration */
				if (pcreq == NULL)
					Caches = creq->next;
				else
					pcreq->next = creq->next;
				free(creq->request);
				free(creq->answer);
				free(creq);
			} else {
				bcopy(creq->answer, ans, creq->anslen);
#ifdef notdef
				/* one time expiration */
				if (pcreq == NULL)
					Caches = creq->next;
				else
					pcreq->next = creq->next;
				free(creq->request);
				free(creq->answer);
				free(creq);
#endif
				return(creq->anslen);
			}
		}
		else if (creq->try >= (u_char)_res.retry) {
			if (LogLevel > 10)
				syslog(LOG_DEBUG, "res_send: cache timeout\n");
			return(-1);     /* timed out */
		}
	} else {
		if (LogLevel > 10)
			show_query("X", buf, buf+buflen);
	}
#endif
	v_circuit = (_res.options & RES_USEVC) || buflen > PACKETSZ;
	badns = 0;

	/*
	 * Send request, RETRY times, or until successful
	 */
	for (try = 0; try < _res.retry; try++) {
	    for (ns = 0; ns < _res.nscount; ns++) {
		struct sockaddr_in *nsap = &_res.nsaddr_list[ns];
    same_ns:
		if (badns & (1<<ns)) {
			_res_close();
			goto next_ns;
		}

#ifdef notdef
		if (Qhook) {
			int done = 0, loops = 0;

			do {
				res_sendhookact act;

				act = (*Qhook)(&nsap,
					       &buf,
					       &buflen,
					       ans,
					       anssiz,
					       &resplen);
				switch (act) {
				case res_goahead:
					done = 1;
					break;
				case res_nextns:
					_res_close();
					goto next_ns;
				case res_done:
					return (resplen);
				case res_modified:
					/* give the hook another try */
					if (++loops < 42) /*doug adams*/
						break;
					/*FALLTHROUGH*/
				case res_error:
					/*FALLTHROUGH*/
				default:
					return (-1);
				}
			} while (!done);
		}
#endif

		Dprint(_res.options & RES_DEBUG,
		       (stdout, ";; Querying server (# %d) address = %s\n",
			ns+1, inet_ntoa(nsap->sin_addr)));

		if (v_circuit) {
			int		truncated;
			struct iovec	iov[2];
			u_short		len;
#if defined(__RES) && (__RES > 19940221)	/* 4.9.3 */
			u_char		*cp;
#else
			char		*cp;
#endif

			/*
			 * Use virtual circuit;
			 * at most one attempt per server.
			 */
			try = _res.retry;
			truncated = 0;
			if ((s < 0) || (!vc)) {
				if (s >= 0)
					_res_close();

				s = socket(AF_INET, SOCK_STREAM, PF_UNSPEC);
				if (s < 0) {
					terrno = errno;
					Perror(stderr, "socket(vc)", errno);
					return (-1);
				}
				if (connect(s,
					    (struct sockaddr *)nsap,
					    sizeof(struct sockaddr))
				    < 0) {
					terrno = errno;
					Aerror(stderr, "connect/vc",
					       errno, *nsap);
					badns |= (1<<ns);
					_res_close();
					goto next_ns;
				}
				vc = 1;
			}
			/*
			 * Send length & message
			 */
			putshort((u_short)buflen, (u_char*)&len);
			iov[0].iov_base = (caddr_t)&len;
			iov[0].iov_len = INT16SZ;
			iov[1].iov_base = (caddr_t)buf;
			iov[1].iov_len = buflen;
			if (writev(s, iov, 2) != (INT16SZ + buflen)) {
				terrno = errno;
				Perror(stderr, "write failed", errno);
				badns |= (1<<ns);
				_res_close();
				goto next_ns;
			}
			/*
			 * Receive length & response
			 */
			cp = ans;
			len = INT16SZ;
			while ((n = read(s, cp, (int)len)) > 0) {
				cp += n;
				if ((len -= n) <= 0)
					break;
			}
			if (n <= 0) {
				terrno = errno;
				Perror(stderr, "read failed", errno);
				_res_close();
				/*
				 * A long running process might get its TCP
				 * connection reset if the remote server was
				 * restarted.  Requery the server instead of
				 * trying a new one.  When there is only one
				 * server, this means that a query might work
				 * instead of failing.  We only allow one reset
				 * per query to prevent looping.
				 */
				if (terrno == ECONNRESET && !connreset) {
					connreset = 1;
					_res_close();
					goto same_ns;
				}
				_res_close();
				goto next_ns;
			}
			resplen = _getshort(ans);
			if (resplen > anssiz) {
				Dprint(_res.options & RES_DEBUG,
				       (stdout, ";; response truncated\n")
				       );
				truncated = 1;
				len = anssiz;
			} else
				len = resplen;
			cp = ans;
			while (len != 0 &&
			       (n = read(s, cp, (int)len)) > 0
			       ) {
				cp += n;
				len -= n;
			}
			if (n <= 0) {
				terrno = errno;
				Perror(stderr, "read(vc)", errno);
				_res_close();
				goto next_ns;
			}
			if (truncated) {
				/*
				 * Flush rest of answer
				 * so connection stays in synch.
				 */
				anhp->tc = 1;
				len = resplen - anssiz;
				while (len != 0) {
					char junk[512];

					n = (len > sizeof(junk)
					     ? sizeof(junk)
					     : len);
					if ((n = read(s, junk, n)) > 0)
						len -= n;
					else
						break;
				}
			}
		} else {
			/*
			 * Use datagrams.
			 */
			struct timeval	timeout;
			fd_set		dsmask;
			struct sockaddr_in from;
			int		fromlen;

			if ((s < 0) || vc) {
				if (vc)
					_res_close();
				s = socket(AF_INET, SOCK_DGRAM, PF_UNSPEC);
				if (s < 0) {
 bad_dg_sock:				terrno = errno;
					Perror(stderr, "socket(dg)", errno);
					return (-1);
				}
				connected = 0;
			}
			/*
			 * On a 4.3BSD+ machine (client and server,
			 * actually), sending to a nameserver datagram
			 * port with no nameserver will cause an
			 * ICMP port unreachable message to be returned.
			 * If our datagram socket is "connected" to the
			 * server, we get an ECONNREFUSED error on the next
			 * socket operation, and select returns if the
			 * error message is received.  We can thus detect
			 * the absence of a nameserver without timing out.
			 * If we have sent queries to at least two servers,
			 * however, we don't want to remain connected,
			 * as we wish to receive answers from the first
			 * server to respond.
			 */
			if (_res.nscount == 1 || (try == 0 && ns == 0)) {
				/*
				 * Connect only if we are sure we won't
				 * receive a response from another server.
				 */
				if (!connected) {
					if (connect(s,
						    (struct sockaddr *)nsap,
						    sizeof(struct sockaddr)
						    ) < 0
					    ) {
						Aerror(stderr,
						       "connect(dg)",
						       errno, *nsap);
						badns |= (1<<ns);
						_res_close();
						goto next_ns;
					}
					connected = 1;
				}
				if (send(s, buf, buflen, 0) != buflen) {
					Perror(stderr, "send", errno);
					badns |= (1<<ns);
					_res_close();
					goto next_ns;
				}
			} else {
				/*
				 * Disconnect if we want to listen
				 * for responses from more than one server.
				 */
				if (connected) {
#if 0 /* defined(BSD) && (BSD >= 199103) */
					struct sockaddr_in no_addr;

					no_addr.sin_family = AF_INET;
					no_addr.sin_addr.s_addr = INADDR_ANY;
					no_addr.sin_port = 0;
					(void) connect(s,
						       (struct sockaddr *)
						        &no_addr,
						       sizeof(no_addr));
#else
					int s1 = socket(AF_INET, SOCK_DGRAM,
							PF_UNSPEC);
					if (s1 < 0)
						goto bad_dg_sock;
					(void) dup2(s1, s);
					(void) close(s1);
					Dprint(_res.options & RES_DEBUG,
					       (stdout, ";; new DG socket\n"))
#endif
					connected = 0;
					errno = 0;
				}
				if (sendto(s, buf, buflen, 0,
					   (struct sockaddr *)nsap,
					   sizeof(struct sockaddr))
				    != buflen) {
					Aerror(stderr, "sendto", errno, *nsap);
					badns |= (1<<ns);
					_res_close();
					goto next_ns;
				}
			}

			/*
			 * Wait for reply
			 */
			timeout.tv_sec = (_res.retrans << try);
			if (try > 0)
				timeout.tv_sec /= _res.nscount;
			if ((long) timeout.tv_sec <= 0)
				timeout.tv_sec = 1;
			timeout.tv_usec = 0;
    wait:
			FD_ZERO(&dsmask);
			FD_SET(s, &dsmask);
			n = select(s+1, &dsmask, (fd_set *)NULL,
				   (fd_set *)NULL, &timeout);
			if (n < 0) {
				Perror(stderr, "select", errno);
				_res_close();
				goto next_ns;
			}
			if (n == 0) {
				/*
				 * timeout
				 */
				Dprint(_res.options & RES_DEBUG,
				       (stdout, ";; timeout\n")
				       );
				gotsomewhere = 1;
				_res_close();
				goto next_ns;
			}
			fromlen = sizeof(struct sockaddr_in);
			resplen = recvfrom(s, ans, anssiz, 0,
					   (struct sockaddr *)&from, &fromlen);
			if (resplen <= 0) {
				Perror(stderr, "recvfrom", errno);
				_res_close();
				goto next_ns;
			}
			gotsomewhere = 1;
			if (hp->id != anhp->id) {
				/*
				 * response from old query, ignore it.
				 * XXX - potential security hazard could
				 *	 be detected here.
				 */
				DprintQ((_res.options & RES_DEBUG) ||
					(_res.pfcode & RES_PRF_REPLY),
					(stdout, ";; old answer:\n"),
					ans);
				goto wait;
			}
#if CHECK_SRVR_ADDR
			if (!(_res.options & RES_INSECURE1) &&
			    !our_server(&from)) {
				/*
				 * response from wrong server? ignore it.
				 * XXX - potential security hazard could
				 *	 be detected here.
				 */
				DprintQ((_res.options & RES_DEBUG) ||
					(_res.pfcode & RES_PRF_REPLY),
					(stdout, ";; not our server:\n"),
					ans);
				goto wait;
			}
#endif
			if (!(_res.options & RES_INSECURE2) &&
			    !queries_match(buf, buf + buflen,
					   ans, ans + anssiz) > 0) {
				/*
				 * response contains wrong query? ignore it.
				 * XXX - potential security hazard could
				 *	 be detected here.
				 */
				DprintQ((_res.options & RES_DEBUG) ||
					(_res.pfcode & RES_PRF_REPLY),
					(stdout, ";; wrong query name:\n"),
					ans);
				goto wait;
			}
			if (anhp->rcode == SERVFAIL ||
			    anhp->rcode == NOTIMP ||
			    anhp->rcode == REFUSED) {
				DprintQ(_res.options & RES_DEBUG,
					(stdout, "server rejected query:\n"),
					ans);
				badns |= (1<<ns);
				_res_close();
				goto next_ns;
			}
			if (!(_res.options & RES_IGNTC) && anhp->tc) {
				/*
				 * get rest of answer;
				 * use TCP with same server.
				 */
				Dprint(_res.options & RES_DEBUG,
				       (stdout, ";; truncated answer\n")
				       );
				v_circuit = 1;
				_res_close();
				goto same_ns;
			}
		} /*if vc/dg*/
		DprintQ((_res.options & RES_DEBUG) ||
			(_res.pfcode & RES_PRF_REPLY),
			(stdout, ";; got answer:\n"),
			ans);
		/*
		 * If using virtual circuits, we assume that the first server
		 * is preferred over the rest (i.e. it is on the local
		 * machine) and only keep that one open.
		 * If we have temporarily opened a virtual circuit,
		 * or if we haven't been asked to keep a socket open,
		 * close the socket.
		 */
		if ((v_circuit && (!(_res.options & RES_USEVC) || ns != 0)) ||
		    !(_res.options & RES_STAYOPEN)) {
			_res_close();
		}
#ifdef notdef
		if (Rhook) {
			int done = 0, loops = 0;

			do {
				res_sendhookact act;

				act = (*Rhook)(nsap,
					       buf,
					       buflen,
					       ans,
					       anssiz,
					       &resplen);
				switch (act) {
				case res_goahead:
				case res_done:
					done = 1;
					break;
				case res_nextns:
					_res_close();
					goto next_ns;
				case res_modified:
					/* give the hook another try */
					if (++loops < 42) /*doug adams*/
						break;
					/*FALLTHROUGH*/
				case res_error:
					/*FALLTHROUGH*/
				default:
					return (-1);
				}
			} while (!done);

		}
#endif
		return (resplen);
    next_ns: ;
	   } /*foreach ns*/
	} /*foreach retry*/
	_res_close();
	if (!v_circuit) {
		if (!gotsomewhere)
			errno = ECONNREFUSED;	/* no nameservers found */
		else
			errno = ETIMEDOUT;	/* no answer obtained */
	} else {
		errno = terrno;
	}
	return (-1);
}

/*
 * This routine is for closing the socket if a virtual circuit is used and
 * the program wants to close it.  This provides support for endhostent()
 * which expects to close the socket.
 *
 * This routine is not expected to be user visible.
 */
void
_res_close()
{
	if (s >= 0) {
		(void) close(s);
		s = -1;
		connected = 0;
		vc = 0;
	}
}

#if RES_NONBLOCK
/*
 * non bocking query routine
 * this routine is called if RES_NONBLOCK flag is set.
 * after of all queries, do res_sync() with RES_NONBLOCK flag.
 */
int
res_cache(buf, buflen, ans, anssiz)
	u_char		*buf;
	int		buflen;
	u_char		*ans;
	int		anssiz;
{
	static int	s = -1;
	static fd_set	dsmask;
	time_t		now, latest;
	struct sockaddr_in *nsap;
	struct timeval	timeout;
	struct sockaddr_in from;
	int		fromlen;
	int		resplen;
	HEADER		*qp, *ap;
	struct cache	*creq;
	int		n;
	int		pending = 0;
	int		syncing = 0;
	int		itvl;

	now = time(NULL);

	if (buf == NULL)
		syncing = 1;
	if (!syncing) {	/* non-block caching */
		for (creq = Caches; creq != NULL; creq = creq->next)
		{
			if (creq->request != NULL
			 && queries_match(buf, buf + buflen, creq->request,
					  creq->request+creq->reqlen) > 0)
				break; /* already registed */
		}

		if (creq == NULL) /* regist in cache */
		{
			if (LogLevel > 10) {
				syslog(LOG_DEBUG, "res_cache: regist in cache\n");
				show_query("Q", buf, buf+buflen);
			}
			creq = (struct cache *)malloc(sizeof(struct cache));
			if (creq == NULL)
				return(-1);	/* allocation failed */
			bzero(creq, sizeof(struct cache));
			creq->request = (u_char *)malloc(buflen);
			if (creq->request == NULL)
				return(-1);	/* allocation failed */
			bcopy(buf, creq->request, buflen);
			creq->reqlen = buflen;
			creq->next = Caches;
			Caches = creq;
		} else {
			if (LogLevel > 10)
				syslog(LOG_DEBUG, "res_cache: found in cache (nop)\n");
			return (0);
		}
	}

	if (s == -1) {
		s = socket(AF_INET, SOCK_DGRAM, PF_UNSPEC);
		if (s < 0) {
			Perror(stderr, "socket", errno);
			return (-1);
		}
	}

again:
	/* send request packets */
	for (creq = Caches; creq != NULL; creq = creq->next)
	{
		if (creq->try >= (u_char)_res.retry)
			continue;	/* timed out */
		if (creq->nocache)
			continue;	/* nocache required */
		if (creq->answer != NULL)
			continue;	/* already got */
		if (creq->time <= now)	/* the time to work */
		{
			if (creq->time != 0)	/* not the first time */
			{
				do {
					++creq->ns;
				} while (creq->badns & (1<<creq->ns));
				if (creq->ns >= (u_char)_res.nscount)
				{
					creq->ns = 0;
					while (creq->badns & (1<<creq->ns))
						creq->ns++;
					if (creq->ns >= (u_char)_res.nscount) {
						/* no valid ns */
						creq->nocache = 1;
						/* set status of timeout */
						creq->try = (u_char)_res.retry;
						if (LogLevel > 10)
							syslog(LOG_DEBUG, "res_cache: no valid ns (%x)\n", creq->badns);
						continue;
					}
					if (creq->try < (u_char)_res.retry)
						creq->try++;
				}
			}
			itvl = _res.retrans << creq->try;
			if (LogLevel > 10) {
				syslog(LOG_DEBUG, "res_cache: itvl %d\n",
					itvl);
			}
			if (creq->try > 0)
				itvl /= _res.nscount;
			if (itvl <= 0)
				itvl = 1;
			creq->time = now + itvl;
			nsap = &_res.nsaddr_list[creq->ns];
			if (LogLevel > 10) {
				syslog(LOG_DEBUG, "res_cache: send query (server is %s)\n", inet_ntoa(nsap->sin_addr));
				show_query("q", creq->request,
					creq->request+creq->reqlen);
			}
			if (sendto(s, creq->request, creq->reqlen, 0,
				   (struct sockaddr *)nsap,
				   sizeof(struct sockaddr)) != creq->reqlen)
			{
				creq->badns |= 1<<creq->ns;
				if (LogLevel > 10)
					syslog(LOG_DEBUG, "res_cache: invalid ns\n");
			}
#if 0 /* HAVE_USLEEP */
			usleep(1000); /* XXX */
#else
			timeout.tv_sec = 0;
			timeout.tv_usec = 1000; /* XXX */
			(void)select(0, (fd_set *)NULL, (fd_set *)NULL,
				     (fd_set *)NULL, &timeout);
#endif
		}
	}

	do {
		latest = 0;

		if (syncing) {
			if (LogLevel > 10)
				syslog(LOG_DEBUG, "res_cache: sync\n");

			/* get next action time */
			for (creq = Caches; creq != NULL; creq = creq->next)
			{
				if (creq->try >= (u_char)_res.retry)
					continue;	/* timed out */
				if (creq->nocache)
					continue;	/* nocache required */
				if (creq->answer != NULL)
					continue;	/* already got */
				if (latest == 0 || creq->time < latest)
					latest = creq->time;
			}

			if (latest == 0)	/* no more pending queries */
				break;
			if (latest > now)
				timeout.tv_sec = latest - now;
			else
				timeout.tv_sec = 1;	/* XXX */
			timeout.tv_usec = 0;
		} else {
			if (LogLevel > 10)
				syslog(LOG_DEBUG, "res_cache: query only\n");

			timeout.tv_sec = 0;
			timeout.tv_usec = 0;
		}

		if (LogLevel > 10)
			syslog(LOG_DEBUG, "res_cache: wait(%d)\n",
				timeout.tv_sec);

		FD_ZERO(&dsmask);
		FD_SET(s, &dsmask);
		if ((n = select(s+1, &dsmask, (fd_set *)NULL, (fd_set *)NULL,
			   &timeout)) > 0)
		{
			/* packet arrived */
			if (LogLevel > 10)
				syslog(LOG_DEBUG, "res_cache: recv answer\n");

			fromlen = sizeof(struct sockaddr_in);
			resplen = recvfrom(s, ans, anssiz, 0,
					   (struct sockaddr *)&from, &fromlen);

			if (LogLevel > 10) {
				/* show_query(ans, ans+anssiz); */
			}

			now = time(NULL);

			if (resplen <= 0) {
				if (LogLevel > 10)
					syslog(LOG_DEBUG, "res_cache: anssiz %d\n", resplen);
				continue;	/* error */
			}

#if CHECK_SRVR_ADDR
			if (!(_res.options & RES_INSECURE1) &&
			    !our_server(&from)) {
				if (LogLevel > 10)
					syslog(LOG_DEBUG, "res_cache: server\n");
				continue;  /* not from our server */
			}
#endif

			ap = (HEADER *)ans;
			for (creq = Caches; creq != NULL; creq = creq->next)
			{
				qp = (HEADER *)creq->request;

				if (qp->id == ap->id
				 && queries_match(ans, ans + resplen,
						  creq->request,
						  creq->request+creq->reqlen)
				    > 0)
				{
					if (ap->tc) {
						Dprint(_res.options & RES_DEBUG, (stdout, ";; truncated answer\n"));
						creq->nocache = 1;
						break;
					}
					if (ap->rcode == SERVFAIL ||
					    ap->rcode == NOTIMP ||
					    ap->rcode == REFUSED) {
						DprintQ(_res.options & RES_DEBUG, (stdout, "server rejected query:\n"), ans);
						creq->badns |= (1<<creq->ns);
						if (LogLevel > 10)
							syslog(LOG_DEBUG, "res_cache: badnsF %x/%x\n", creq->badns, (1<<_res.nscount)-1);
						if (creq->badns != (1<<_res.nscount)-1)
							break;
					}

					if (LogLevel > 10) {
						syslog(LOG_DEBUG, "res_cache: cache\n");
						show_query("A", ans, ans+resplen);
					}

					prefetch_related(ans, ans+resplen);

					creq->answer = (u_char*)malloc(resplen);
					if (creq->answer != NULL) {
						bcopy(ans, creq->answer,
						resplen);
					} else {
						/* error */
					}
					creq->anslen = resplen;
					creq->try = (u_char)_res.retry;
					creq->time = now + 3600; /* XXX */

					break;
				}
			}
			if (creq == NULL)
			{
				/* no associated query */
				if (LogLevel > 10)
					syslog(LOG_DEBUG, "res_cache: what??\n");
				continue;
			}
		}
		if (n < 0) {
			/* error in select */
			break;
		}
	} while(n > 0);

	if (syncing) {
		pending = 0;
		if (LogLevel > 10)
			syslog(LOG_DEBUG, "res_cache: syncing (%d)\n", now);
		for (creq = Caches; creq != NULL; creq = creq->next)
		{
			if (creq->try >= (u_char)_res.retry)
				continue;	/* timed out */
			if (creq->nocache)
				continue;	/* nocache required */
			if (creq->answer != NULL)
				continue;

			pending++;
			if (LogLevel > 10)
				syslog(LOG_DEBUG, "ev: %d\n", creq->time);
		}
		if (pending)
		{
			if (LogLevel > 10)
				syslog(LOG_DEBUG, "res_cache: sync retry\n");
			/* sleep(1); */
			now = time(NULL);
			goto again;
		}
	}
	return (0);
}

/*
 * for sync operation
 */
/* void */
res_sync()
{
        char qbuf[MAXDNAME];
	(void)res_send(NULL, 0, qbuf, MAXDNAME);
}


prefetch_related(b, e)
u_char *b, *e;
{
	char nbuf[MAXDNAME+1];
	register int n;
	int type, class, pref;
	HEADER *hp;
	u_char *eom, *ap;
	struct hostent *hep;
	int ancount, qdcount;
        union
        {
                HEADER  qb1;
                char    qb2[PACKETSZ];
        } answer;

	hp = (HEADER *) b;
	ap = (u_char *) b + sizeof(HEADER);
	eom = (u_char *) e;

	/* skip question part of response -- we know what we asked */
	for (qdcount = ntohs(hp->qdcount); qdcount--; )
	{
		n = dn_expand((u_char *) b, eom, ap,
			      (u_char *) nbuf, sizeof nbuf);
		ap += n;
		GETSHORT(type, ap);
		GETSHORT(class, ap);
	}

	if (type != T_MX)
	{
		return;
	}

	if (ntohs(hp->ancount) == 0)
	{
		res_query(nbuf, C_IN, T_A, (char *)&answer, sizeof(answer));
		return;
	}

	for (ancount = ntohs(hp->ancount); --ancount >= 0 && ap < eom;)
	{
		n = dn_expand((u_char *) b, eom, ap,
			      (u_char *) nbuf, sizeof nbuf);
		if (n < 0)
			break;
		ap += n;
		GETSHORT(type, ap);
		GETSHORT(class, ap);
		ap += sizeof(long);
		GETSHORT(n, ap);
		if (type == T_MX) {
			GETSHORT(pref, ap);
			if ((n = dn_expand((u_char *) b, eom, ap,
					   (u_char *) nbuf, sizeof nbuf)) < 0)
				break;
			ap += n;
			res_query(nbuf, C_IN, T_A,
					(char *)&answer, sizeof(answer));
		} else {
			ap += n;
		}
	}
}

show_query(t, b, e)
u_char *t, *b, *e;
{
	char nbuf[MAXDNAME+1];
	register int n;
	int type, class, pref;
	HEADER *hp;
	u_char *eom, *ap;
	struct hostent *hep;
	int ancount, qdcount;

	hp = (HEADER *) b;
	ap = (u_char *) b + sizeof(HEADER);
	eom = (u_char *) e;

	/* skip question part of response -- we know what we asked */
	for (qdcount = ntohs(hp->qdcount); qdcount--; )
	{
		n = dn_expand((u_char *) b, eom, ap,
			      (u_char *) nbuf, sizeof nbuf);
		ap += n;
		GETSHORT(type, ap);
		GETSHORT(class, ap);
		syslog(LOG_DEBUG, "%s:%s(%d,%d), rcode=%d\n",
			t, nbuf, class, type, ntohs(hp->rcode));
	}

	if (ntohs(hp->ancount) == 0)
		return;

	syslog(LOG_DEBUG, "ancount = %d\n", ntohs(hp->ancount));
	for (ancount = ntohs(hp->ancount); --ancount >= 0 && ap < eom;)
	{
		n = dn_expand((u_char *) b, eom, ap,
			      (u_char *) nbuf, sizeof nbuf);
		if (n < 0)
			break;
		ap += n;
		GETSHORT(type, ap);
		GETSHORT(class, ap);
		ap += sizeof(long);
		GETSHORT(n, ap);
		if (type == T_MX) {
			GETSHORT(pref, ap);
			if ((n = dn_expand((u_char *) b, eom, ap,
					   (u_char *) nbuf, sizeof nbuf)) < 0)
				break;
			ap += n;
			syslog(LOG_DEBUG, "MX pref=%d,host=(%s)\n", pref, nbuf);
#if 0
			hep = gethostbyname(nbuf);
			if (hep == NULL) {
				syslog(LOG_DEBUG, "get A --  h_errno=%d\n",
					h_errno);
			} else {
				syslog(LOG_DEBUG, "get A -- OK\n");
			}
#endif
		} else {
			ap += n;
			/* syslog(LOG_DEBUG, "unexpected answer type: %d\n",
				type); */
		}
	}
}
#endif
